package indexed;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Random;

/**
 * The indexed list class.
 * @author Ronald Mak
 */
public class IndexedList implements List<Integer>
{
    /** The internal index array that points to every kth indexed list node. */
    private ArrayList<Node> index = null; 
    
    private int k = 1;          // index array points to every kth node
    private int xPosition;      // a position of an index array element
    private int xDelta;         // no. of nodes to move forward or backward
                                //   from an indexed node
    private int xResidual = 0;  // no. of nodes following the last indexed node
    
    private Node head = null;   // indexed list header node
    private Node tail = null;   // indexed list tail node
    private int N = 0;          // number of data nodes in the indexed list
    
    /**
     * Constructor.
     * @param k the index array points to every kth indexed list node.
     */
    public IndexedList(int k)
    {
        clear();
        this.k = k;
    }
    
    /**
     * Constructor.
     * @param values the values to enter into the indexed list
     * @param k the index array points to every kth indexed list node.
     */
    public IndexedList(ArrayList<Integer> values, int k)
    {
        this(k);
        int m = 0;  // counter for every kth node
        
        for (int i = 0; i < values.size(); i++) {
            Node node = append(values.get(i));
            
            // The index array points to every kth node beginning with the 0th.
            if (m-- == 0) {
                index.add(node);
                m = k-1;
            }
        }
        
        // Number of nodes following the last one pointed to by the index array.
        xResidual = N - k*(index.size() - 1) - 1;
    }
    
    //////////////////////
    //                  //
    // Key List methods //
    //                  //
    //////////////////////
    
    /**
     * Reset the indexed list to "empty".
     */
    @Override
    public void clear()
    {
        this.head = new Node(-Integer.MIN_VALUE, null, null);
        this.tail = new Node(+Integer.MAX_VALUE, this.head, null);
        this.head.setNext(this.tail);
        this.N = 0;
        
        this.index = new ArrayList<Node>();
        this.xResidual = 0;
    }
    
    /**
     * Return the number of data nodes in the indexed list.
     * @return the number of nodes.
     */
    @Override
    public int size()
    {
        return N;
    }
    
    /**
     * Get the data at the ith position of the indexed list.
     * @param i the index.
     * @return the data.
     */
    @Override
    public Integer get(int i)
    {
        return getNode(i).data();
    }
    
    /**
     * Set the data at the ith position of the indexed list.
     * @param i the index.
     * @param data the data to set.
     * @return the data.
     */
    @Override 
    public Integer set(int i, Integer data) 
    { 
        Node node = getNode(i);
        node.setData(data);
        
        return data; 
    }
    
    /**
     * Insert a new node at the ith position of the indexed list.
     * @param i the index.
     * @param data the data for the new node.
     */
    @Override
    public void add(int i, Integer data)
    {
        Node nodeAfter = getNode(i);
        Node newNode = new Node(data);
        insert(newNode, nodeAfter);
        
        // Initialize the index array if this is the first indexed list node.
        if (N == 1) {
            index.add(newNode);
            xResidual = 0;
        }
        
        // Update the index array by pointing backward one node
        // following the insertion point.
        else if (xDelta > 0) {
            updateIndexBackward(xPosition + 1);
        }
        else {
            updateIndexBackward(xPosition);
        }
    }
    
    /**
     * Delete the node at the ith position of the indexed list.
     * @param i the index.
     * @return the data of the deleted node.
     */
    @Override
    public Integer remove(int i)
    {
        Node node = getNode(i);        
        delete(node);
        
        // Update the index array by pointing forward one node
        // following the insertion point.
        if (xDelta > 0) {
            updateIndexForward(xPosition + 1);
        }
        else {
            updateIndexForward(xPosition);
        }
        
        return node.data();
    }

    //////////////////////////////
    //                          //
    // Internal private methods //
    //                          //
    //////////////////////////////
    
    /**
     * Return the node at the ith position of the indexed list
     * as accessed via the index array.
     * @param i the position in the indexed list.
     * @return the ith node.
     */
    private Node getNode(int i)
    {
        // Compute the position of the index array (xPosition) whose element
        // points to or closest to the ith node of the indexed array.
        // Compute the number of nodes (xDelta) to get to the ith node
        // from the node (the indexed node) pointed to by the index array:
        //   xDelta > 0: move forward 
        //   xDelta < 0: move backward
        xPosition = i/k;
        xDelta = i - k*xPosition;
        
        // If more than halfway forward from the indexed node, use the next one.
        if (xDelta > k/2) {
            if (xPosition < index.size()-1) {
                xPosition++;
                xDelta -= k;  // now negative to move backward
            }
        }
        
        // Get the indexed node.
        Node node = index.get(xPosition);
        
        // Move forward from the indexed node.
        if (xDelta > 0) {
            int d = xDelta;
            while (d-- > 0) node = node.next();
        }
        
        // Move backward from the indexed node.
        else {
            int d = xDelta;
            while (d++ < 0) node = node.prev();
        }
        return node;
    }
    
    /**
     * Append a new node to the end of the indexed list.
     * Does not update the index array.
     * @param newNode the new node.
     * @return the new node.
     */
    private Node append(Node newNode)
    {
        newNode.setPrev(tail.prev());
        newNode.setNext(tail);
        tail.prev().setNext(newNode);
        tail.setPrev(newNode);
        
        N++;
        return newNode;
    }
    
    /**
     * Append a new value to the end of the indexed list.
     * Does not update the index array.
     * @param value the new value.
     * @return the new node.
     */
    private Node append(int value)
    {
        Node newNode = new Node(value);
        return append(newNode);
    }
    
    /**
     * Insert a new node into the indexed list before an existing node.
     * Does not update the index array.
     * @param newNode the new node.
     * @param nodeAfter the node that will come after the new node.
     * @return the new node.
     */
    private Node insert(Node newNode, Node nodeAfter)
    {
        newNode.setPrev(nodeAfter.prev());
        newNode.setNext(nodeAfter);
        nodeAfter.prev().setNext(newNode);  
        nodeAfter.setPrev(newNode);
        
        N++;
        return newNode;
    }
    
    /**
     * Insert a new value into the indexed list before an existing node.
     * Does not update the index array.
     * @param value the new value.
     * @param nodeAfter the node that will come after the new node.
     * @return the new node.
     */
    private Node insert(int value, Node beforeNode)
    {
        Node newNode = new Node(value);
        return insert(newNode, beforeNode);
    }
    
    /**
     * Delete a node from the indexed list.
     * Does not update the index array.
     * @param node the node to delete.
     * @return the deleted node.
     */
    private Node delete(Node node)
    {
        node.prev().setNext(node.next());
        node.next().setPrev(node.prev());
        
        N--;
        return node;
    }
    
    /**
     * Update the index array after an insertion into the indexed list
     * by pointing each element of the array to the previous node,
     * starting after the insertion point.
     * @param pos the starting position of the index array to update.
     */
    private void updateIndexBackward(int pos)
    {
        for (int j = pos; j < index.size(); j++) {
            index.set(j, index.get(j).prev());
        }
        
        // If the residual is now k, append a new index array element.
        if (++xResidual == k) {
            index.add(tail.prev());
            xResidual = 0;
        }
    }
    
    /**
     * Update the index array after a deletion from the indexed list
     * by pointing each element of the array to the next node,
     * starting after the deletion point.
     * @param pos the starting position of the index array to update.
     */
    private void updateIndexForward(int pos)
    {
        for (int j = pos; j < index.size(); j++) {
            index.set(j, index.get(j).next());
        }
        
        // If the residual was 0, delete the last index array element.
        if (xResidual-- == 0) {
            index.remove(index.get(index.size()-1));
            xResidual = k-1;
        }
    }
    
    /**
     * Return the first node of the indexed list.
     * @return the first node.
     */
    private Node first()
    {
        return head.next();
    }
    
    /**
     * Return the last node of the indexed list.
     * @return the last node.
     */
    private Node last()
    {
        return tail.prev();
    }

    /**
     * Print the data in the indexed list forwards and backwards.
     * Does not use the index array.
     * @param label the label to precede the data.
     */
    private void printData(String label)
    {
        // Print the data forwards.
        System.out.print(label + "  (forwards):");
        for (Node node = head.next(); node != tail; node = node.next()) {
            System.out.print(" " + node.data());
        }
        System.out.println();
        
        // Print the data backwards.
        System.out.print(label + " (backwards):");
        for (Node node = tail.prev(); node != head; node = node.prev()) {
            System.out.print(" " + node.data());
        }
        System.out.println();
    }
    
    /**
     * Print the data in the indexed list as accessed via the index array.
     * Also print the index array.
     */
    private void dump()
    {
        // Print the indexed list data.
        System.out.print("List:");
        for (int i = 0; i < N; i++) {
            Integer value = get(i);
            System.out.print(" " + value);
        }
        System.out.println();

        // Print the index array.
        System.out.print("Index:");
        for (int j = 0; j < index.size(); j++) {
            System.out.printf(" %d:%d", j, index.get(j).data());
        }
        System.out.println();
        System.out.println("Residual=" + xResidual);
    }
    
    //////////
    //      //
    // Main //
    //      //
    //////////
    
    /**
     * The main for testing the indexed list.
     * @param args the command-line arguments (ignored).
     */
    public static void main(String args[])
    {
        IndexedList list = new IndexedList(10);

        for (int i = 0; i < 10; i += 2) {
            list.append(i);
        }
        list.printData("Even appends");
        
        int val = 1;
        Node node = list.first().next();
        while (node != list.last().next()) {
            list.insert(val, node);
            val += 2;
            node = node.next();
        }
        list.printData("Odd inserts");
        
        node = list.first();
        do {
            Node nextNode = node.next();
            list.delete(node);
            node = nextNode.next();
        } while (node != null);
        list.printData("Even deletes");
        
        int n = 30;
        ArrayList<Integer> values = new ArrayList<>();
        
        for (int i = 0; i < n; i++) {
            values.add(100*i);
        }

        list = new IndexedList(values, 10);
        list.printData("Initialized");
        
        System.out.println();
        list.dump();

        System.out.println();
        System.out.println("Adding 0 @ 0");
        list.add(0, 0);
        list.dump();
        
        System.out.println();
        System.out.println("Removing @ 0");
        list.remove(0);
        list.dump();
        
        for (int value = 1299; value >= 1270; value--) {
            int index = 13;
            System.out.println();
            System.out.println("Adding " + value + " @ " + index);
            list.add(index, value);
            list.dump();
        }
        
        for (int value = 899; value >= 870; value--) {
            int index = 9;
            System.out.println();
            System.out.println("Adding " + value + " @ " + index);
            list.add(index, value);
            list.dump();
        }
        
        for (int value = 870; value <= 899; value++) {
            int index = 9;
            System.out.println();
            System.out.println("Deleting " + value + " @ " + index);
            list.remove(index);
            list.dump();
        }
        
        for (int value = 1270; value <= 1299; value++) {
            int index = 13;
            System.out.println();
            System.out.println("Deleting " + value + " @ " + index);
            list.remove(index);
            list.dump();
        }
        
        Random r = new Random();
        
        System.out.println();
        for (int m = 1; m < n; m++) {
            int i = r.nextInt(list.size()-1);
            System.out.println("Adding " + -i + " @ " + i);
            list.add(i, -i);
        }
        System.out.println();
        list.dump();
        
        System.out.println();
        for (int m = 1; m < n; m++) {
            int i = r.nextInt(list.size()-1);
            System.out.println("Removing @ " + i);
            list.remove(i);
        }
        System.out.println();
        list.dump();
    }
    
    ////////////////////////////
    //                        //
    // Remaining List methods //
    //                        //
    ////////////////////////////
    
    @Override public boolean add(Integer value) { return false; }
    @Override public boolean addAll(Collection<? extends Integer> c) { return false; }
    @Override public boolean addAll(int index, Collection<? extends Integer> c) { return false; }
    @Override public boolean contains(Object o) { return false; }
    @Override public boolean containsAll(Collection<?> c) { return false; }
    @Override public int indexOf(Object o) { return 0; }
    @Override public boolean isEmpty() { return false; }
    @Override public Iterator<Integer> iterator() { return null; }
    @Override public int lastIndexOf(Object o) { return 0; }
    @Override public ListIterator<Integer> listIterator() {  return null; }
    @Override public ListIterator<Integer> listIterator(int index) { return null; }
    @Override public boolean remove(Object o) { return false; }
    @Override public boolean removeAll(Collection<?> c) {  return false; }
    @Override public boolean retainAll(Collection<?> c) { return false; }
    @Override public List<Integer> subList(int fromIndex, int toIndex) { return null; }
    @Override public Object[] toArray() { return null; }
    @Override public <T> T[] toArray(T[] a) { return null; }
}