package indexed;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

public class IndexedTimings 
{
    private static ArrayList<Integer>  arrayList = null;
    private static LinkedList<Integer> linkedList = null;
    private static IndexedList indexedList = null;
    
    /**
     * Initialize the array list, linked list, and the indexed list to all zeros. 
     * @param n the list size N.
     * @param k the value of k.
     */
    private static void initLists(int n, int k)
    {
        arrayList = new ArrayList<>();
        
        for (int i = 0; i < n; i++) {
            arrayList.add(0);
        }
        
        linkedList  = new LinkedList<>(arrayList);
        indexedList = new IndexedList(arrayList, k);
    }
    
    /**
     * Time the get() calls to each node of the list.
     * @param list the list.
     * @return the time in ms.
     */
    private static long timeGet(List<Integer> list)
    {
        long start = System.currentTimeMillis();
        
        for (int i = 0; i < list.size(); i++) {
            list.get(i);
        }
        
        return System.currentTimeMillis() - start;
    }
    
    /**
     * Time the add() calls to the list. 
     * Add to random indexes until the list has doubled its starting size.
     * @param list the list.
     * @return the time in ms.
     */
    private static long timeAdd(List<Integer> list)
    {
        int m = list.size();
        Random r = new Random();
        
        long start = System.currentTimeMillis();
        
        for (int j = 0; j < m; j++) {
            int i = r.nextInt(list.size());
            list.add(i, 0);
        }
        
        return System.currentTimeMillis() - start;
    }
    
    /**
     * Time the remove() calls to the list.
     * Delete from random indexes until the list is half its starting size.
     * @param list the list.
     * @return the time in ms.
     */
    private static long timeRemove(List<Integer> list)
    {
        int m = list.size()/2;
        Random r = new Random();
        
        long start = System.currentTimeMillis();
        
        for (int j = 0; j < m; j++) {
            int i = r.nextInt(list.size());
            list.remove(i);
        }
        
        return System.currentTimeMillis() - start;
    }
    
    /**
     * Time a mixture of get(), add(), and remove() calls to the list.
     * Make as many calls as half the starting size of the list.
     * @param list the list.
     * @return the time in ms.
     */
    private static long timeMixture(List<Integer> list)
    {
        Random r1 = new Random();  // random list index
        Random r2 = new Random();  // random call
        int m = list.size()/2;
        
        long start = System.currentTimeMillis();
        
        for (int j = 0; j < m; j++) {
            int i = r1.nextInt(list.size());
            
            switch (r2.nextInt(2)) {
                case 0: {
                    list.get(i);
                    break;
                }
                case 1: {
                    list.add(i, 0);
                    break;
                }
                case 2: {
                    list.remove(i);
                    break;
                }
            }
        }
        
        return System.currentTimeMillis() - start;
    }
    
    /**
     * Time get() calls to all three lists.
     */
    private static void testGet()
    {
        int k = 100;
        
        System.out.printf("get(i), k = %d\n\n", k);
        System.out.printf("%8s%12s%12s%12s\n", "N", 
                          "ArrayList", "LinkedList", "IndexedList");
        
        for (int n = 100; n <= 100000; n*=10) {
            initLists(n, k);
            
            long timeArray  = timeGet(arrayList);
            long timeLinked = timeGet(linkedList);
            long timeIndexed = timeGet(indexedList);
            
            System.out.printf("%8d%12d%12d%12d\n", 
                              n, timeArray, timeLinked, timeIndexed);
        }
    }
    
    /**
     * Time add() calls to all three lists.
     * Start with a list of size N/2 and grow to size N.
     */
    private static void testAdd()
    {
        int k = 100;
        
        System.out.printf("add(x,i), k = %d\n\n", k);
        System.out.printf("%8s%12s%12s%12s\n", "N", 
                          "ArrayList", "LinkedList", "IndexedList");
        
        for (int n = 100; n <= 100000; n*=10) {
            initLists(n/2, k);
            
            long timeArray  = timeAdd(arrayList);
            long timeLinked = timeAdd(linkedList);
            long timeIndexed = timeAdd(indexedList);
            
            System.out.printf("%8d%12d%12d%12d\n", 
                              n, timeArray, timeLinked, timeIndexed);
        }
    }

    /**
     * Time remove() calls to all three lists.
     * Start with a list of N and shrink to size N/2.
     */
    private static void testRemove()
    {
        int k = 100;
        
        System.out.printf("remove(i), k = %d\n\n", k);
        System.out.printf("%8s%12s%12s%12s\n", "N", 
                          "ArrayList", "LinkedList", "IndexedList");
        
        for (int n = 100; n <= 100000; n*=10) {
            initLists(n, k);
            
            long timeArray  = timeRemove(arrayList);
            long timeLinked = timeRemove(linkedList);
            long timeIndexed = timeRemove(indexedList);
            
            System.out.printf("%8d%12d%12d%12d\n", 
                              n, timeArray, timeLinked, timeIndexed);
        }
    }
    
    /**
     * Time a mixture of get(), add(), and remove() calls to all three lists.
     */
    private static void testMixture()
    {
        int k = 100;
        
        System.out.printf("Mixture, k = %d\n\n", k);
        System.out.printf("%8s%12s%12s%12s\n", "N", 
                          "ArrayList", "LinkedList", "IndexedList");
        
        for (int n = 100; n <= 100000; n*=10) {
            initLists(n, k);
            
            long timeArray  = timeMixture(arrayList);
            long timeLinked = timeMixture(linkedList);
            long timeIndexed = timeMixture(indexedList);
            
            System.out.printf("%8d%12d%12d%12d\n", 
                              n, timeArray, timeLinked, timeIndexed);
        }
    }
    
    /** Values of k to try */
    private static final int K[] = {2, 10, 100, 1000, 10000};
    
    /**
     * Time get(), add(), remove(), and a mixture of calls to the indexed list
     * with various values of k.
     */
    private static void timeK()
    {
        int n = 100000;
        
        System.out.printf("Indexed list, N = %d\n\n", n);
        System.out.printf("%9s", "Operation");
        for (int k = 0; k < K.length; k++) {
            System.out.printf("%8s", "k=" + K[k]);
        }
        System.out.println();
        
        System.out.printf("%9s", "get(i)");
        for (int k = 0; k < K.length; k++) {
            initLists(n, K[k]);
            long timeIndexed = timeGet(indexedList);
            System.out.printf("%8d", timeIndexed);
        }
        System.out.println();
        
        System.out.printf("%9s", "add(x,i)");
        for (int k = 0; k < K.length; k++) {
            initLists(n/2, K[k]);
            long timeIndexed = timeAdd(indexedList);
            System.out.printf("%8d", timeIndexed);
        }
        System.out.println();
        
        System.out.printf("%9s", "remove(i)");
        for (int k = 0; k < K.length; k++) {
            initLists(n, K[k]);
            long timeIndexed = timeRemove(indexedList);
            System.out.printf("%8d", timeIndexed);
        }
        System.out.println();
        
        System.out.printf("%9s", "mixture");
        for (int k = 0; k < K.length; k++) {
            initLists(n, K[k]);
            long timeIndexed = timeMixture(indexedList);
            System.out.printf("%8d", timeIndexed);
        }
        System.out.println();
    }
    
    /**
     * The main for running the timing tests.
     * @param args the command-line arguments (ignored).
     */
    public static void main(String[] args) 
    {
        testGet();
        System.out.println();
        
        testAdd();
        System.out.println();
        
        testRemove();
        System.out.println();
        
        testMixture();
        System.out.println();
        
        timeK();
    }
}
